/*
 * %CopyrightBegin%
 *
 * Copyright Ericsson AB 2008-2016. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * %CopyrightEnd% 
*/
/***** This file is generated do not edit ****/

static struct {
   const char * name;
   const char * alt;
   void * func;
} glu_fns[] = 
{
 {"gluBuild1DMipmapLevels", NULL, &wegluBuild1DMipmapLevels},
 {"gluBuild1DMipmaps", NULL, &wegluBuild1DMipmaps},
 {"gluBuild2DMipmapLevels", NULL, &wegluBuild2DMipmapLevels},
 {"gluBuild2DMipmaps", NULL, &wegluBuild2DMipmaps},
 {"gluBuild3DMipmapLevels", NULL, &wegluBuild3DMipmapLevels},
 {"gluBuild3DMipmaps", NULL, &wegluBuild3DMipmaps},
 {"gluCheckExtension", NULL, &wegluCheckExtension},
 {"gluCylinder", NULL, &wegluCylinder},
 {"gluDeleteQuadric", NULL, &wegluDeleteQuadric},
 {"gluDisk", NULL, &wegluDisk},
 {"gluErrorString", NULL, &wegluErrorString},
 {"gluGetString", NULL, &wegluGetString},
 {"gluLookAt", NULL, &wegluLookAt},
 {"gluNewQuadric", NULL, &wegluNewQuadric},
 {"gluOrtho2D", NULL, &wegluOrtho2D},
 {"gluPartialDisk", NULL, &wegluPartialDisk},
 {"gluPerspective", NULL, &wegluPerspective},
 {"gluPickMatrix", NULL, &wegluPickMatrix},
 {"gluProject", NULL, &wegluProject},
 {"gluQuadricDrawStyle", NULL, &wegluQuadricDrawStyle},
 {"gluQuadricNormals", NULL, &wegluQuadricNormals},
 {"gluQuadricOrientation", NULL, &wegluQuadricOrientation},
 {"gluQuadricTexture", NULL, &wegluQuadricTexture},
 {"gluScaleImage", NULL, &wegluScaleImage},
 {"gluSphere", NULL, &wegluSphere},
 {"gluUnProject", NULL, &wegluUnProject},
 {"gluUnProject4", NULL, &wegluUnProject4},
 { NULL, NULL, NULL}};
