%%
%% %CopyrightBegin%
%%
%% Copyright Ericsson AB 2018. All Rights Reserved.
%%
%% Licensed under the Apache License, Version 2.0 (the "License");
%% you may not use this file except in compliance with the License.
%% You may obtain a copy of the License at
%%
%%     http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing, software
%% distributed under the License is distributed on an "AS IS" BASIS,
%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
%% See the License for the specific language governing permissions and
%% limitations under the License.
%%
%% %CopyrightEnd%

-module(cdv_persistent_cb).

-export([get_info/0]).

-include_lib("wx/include/wx.hrl").

get_info() ->
    Tab = ets:new(pt_expand,[set,public]),
    {ok,PT,TW} = crashdump_viewer:persistent_terms(),
    {{expand,
      observer_html_lib:expandable_term("Persistent Terms",PT,Tab),
      Tab},
     TW}.
